==> gossiplog-0.14.0-next.2.tgz: package/lib/AbstractGossipLog.js <==
import { TypedEventEmitter } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { equals, toString } from "uint8arrays";
import { hashEntry } from "@canvas-js/okra";
import { ed25519, prepareMessage } from "@canvas-js/signatures";
import { assert, zip } from "@canvas-js/utils";
import { AbortError, MessageNotFoundError, MissingParentError } from "@canvas-js/gossiplog/errors";
import * as sync from "@canvas-js/gossiplog/sync";
import target from "#target";
import { namespace, version } from "#migrations";
import { AncestorIndex } from "./AncestorIndex.js";
import { SignedMessage } from "./SignedMessage.js";
import { MessageSet } from "./MessageSet.js";
import { MessageId, decodeId, encodeId, messageIdPattern } from "./MessageId.js";
import { getNextClock } from "./schema.js";
import { gossiplogTopicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    db;
    tree;
    static namespace = namespace;
    static version = version;
    static schema = {
        $messages: {
            id: "primary",
            data: "bytes",
            hash: "string",
            clock: "integer",
            $indexes: ["clock"],
        },
        $heads: { id: "primary" },
        $replays: {
            timestamp: "primary",
            cursor: "string?",
            $indexes: ["cursor"],
        },
        ...AncestorIndex.schema,
    };
    topic;
    signer;
    controller = new AbortController();
    log;
    ancestorIndex;
    validatePayload;
    verifySignature;
    #apply;
    constructor(db, tree, init) {
        super();
        this.db = db;
        this.tree = tree;
        assert(gossiplogTopicPattern.test(init.topic), "invalid topic (must be of the form 'topic' or 'topic#hash', where topic matches [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.signer = init.signer ?? ed25519.create();
        this.#apply = init.apply;
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
        this.ancestorIndex = new AncestorIndex(db);
    }
    async initialize() {
        const replays = await this.db.query("$replays", {
            where: { cursor: { neq: null } },
            orderBy: { timestamp: "desc" },
        });
        if (replays.length > 0) {
            const [{ timestamp, cursor }, ...rest] = replays;
            assert(cursor !== null, "internal error - expected cursor !== null");
            this.log("found incomplete replay from %s at cursor %s", timestamp, cursor);
            // clear existing replay records
            await this.db.apply(rest.map(({ timestamp }) => ({
                model: "$replays",
                operation: "set",
                value: { timestamp, cursor: null },
            })));
            await this.#replay(timestamp, cursor);
        }
    }
    async close() {
        this.log("closing");
        this.controller.abort();
        await new Promise((resolve) => setTimeout(resolve, 0));
        await this.tree.close();
        await this.db.close();
    }
    async replay() {
        this.log("beginning replay");
        // clear existing replay records
        await this.db.query("$replays", { where: { cursor: { neq: null } } }).then(async (replays) => {
            const effects = replays.map((replay) => ({
                model: "$replays",
                operation: "set",
                value: { ...replay, cursor: null },
            }));
            await this.db.apply(effects);
        });
        const timestamp = new Date().toISOString();
        await this.db.set("$replays", { timestamp, cursor: null });
        for (const name of ["$heads", ...Object.keys(AncestorIndex.schema)]) {
            this.log("clearing model %s", name);
            await this.db.clear(name);
        }
        await this.tree.clear();
        return await this.#replay(timestamp);
    }
    async #replay(timestamp, cursor) {
        const pageSize = 128;
        const lowerBound = { gt: cursor };
        while (!this.controller.signal.aborted) {
            this.log("fetching new page");
            const results = await this.db.query("$messages", {
                orderBy: { id: "asc" },
                where: { id: lowerBound },
                limit: pageSize,
            });
            this.log("got new page of %d records", results.length);
            if (results.length === 0) {
                this.log("finished replaying");
                await this.db.set("$replays", { timestamp, cursor: null });
                return true;
            }
            let [{ id: cursor }] = results;
            await this.tree.write(async (txn) => {
                for (const { id, data } of results) {
                    this.log("replaying message %s", id);
                    const signedMessage = this.decode(data);
                    assert(signedMessage.id === id, "internal error - expected signedMessage.id === id");
                    await this.apply(txn, signedMessage);
                    cursor = id;
                    if (this.controller.signal.aborted) {
                        break;
                    }
                }
            });
            lowerBound.gt = cursor;
            await this.db.set("$replays", { timestamp, cursor });
        }
        this.log("replay incomplete");
        return false;
    }
    async connect(url, options = {}) {
        return await target.connect(this, url, options);
    }
    async listen(port, options = {}) {
        await target.listen(this, port, options);
    }
    async startLibp2p(config) {
        const libp2p = await target.startLibp2p(this, config);
        this.controller.signal.addEventListener("abort", () => libp2p.stop());
        return libp2p;
    }
    encode(signature, message, context = {}) {
        if (this.topic !== message.topic) {
            this.log.error("invalid message: %O", message);
            throw new Error(`cannot encode message: expected topic ${this.topic}, found topic ${message.topic}`);
        }
        const preparedMessage = prepareMessage(message);
        if (!this.validatePayload(preparedMessage.payload)) {
            this.log.error("invalid message: %O", preparedMessage);
            throw new Error(`error encoding message: invalid payload`);
        }
        return SignedMessage.encode(signature, preparedMessage, context);
    }
    decode(value, context = {}) {
        const signedMessage = SignedMessage.decode(value, context);
        const { topic, payload } = signedMessage.message;
        if (this.topic !== topic) {
            this.log.error("invalid message: %O", signedMessage.message);
            throw new Error(`cannot decode message: expected topic ${this.topic}, found topic ${topic}`);
        }
        if (!this.validatePayload(payload)) {
            this.log.error("invalid message: %O", signedMessage.message);
            throw new Error(`error decoding message: invalid payload`);
        }
        return signedMessage;
    }
    async getClock() {
        const heads = await this.db.getAll("$heads");
        const ids = heads.map(({ id }) => id);
        const clock = getNextClock(ids.map(encodeId));
        return [clock, ids];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const records = await this.db.query("$messages", { select: { id: true }, where: { id } });
        return records.length > 0;
    }
    async get(id) {
        const record = await this.db.get("$messages", id);
        if (record === null) {
            return null;
        }
        else {
            return this.decode(record.data);
        }
    }
    async getMessages(range = {}) {
        const { reverse = false, limit, ...where } = range;
        const records = await this.db.query("$messages", {
            where: { id: where },
            select: { id: true, data: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
        return records.map(({ data }) => this.decode(data));
    }
    async *iterate(range = {}) {
        const { reverse = false, limit, ...where } = range;
        for await (const row of this.db.iterate("$messages", {
            where: { id: where },
            select: { id: true, data: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        })) {
            yield this.decode(row.data);
        }
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The current concurrent heads of the local log are used as parents.
     */
    async append(payload, { signer = this.signer } = {}) {
        let root = null;
        let heads = null;
        let result = undefined;
        const signedMessage = await this.tree.write(async (txn) => {
            const [clock, parents] = await this.getClock();
            const message = prepareMessage({ topic: this.topic, clock, parents, payload });
            const signature = await signer.sign(message);
            const signedMessage = this.encode(signature, message);
            this.log("appending message %s at clock %d with parents %o", signedMessage.id, clock, parents);
            const applyResult = await this.apply(txn, signedMessage);
            root = applyResult.root;
            heads = applyResult.heads;
            result = applyResult.result;
            return signedMessage;
        });
        assert(root !== null && heads !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root, heads } }));
        signedMessage.result = result;
        return signedMessage;
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(signedMessage) {
        const { message, signature } = signedMessage;
        assert(message.topic === this.topic, `expected message.topic === this.topic`);
        await this.verifySignature(signature, message);
        const { clock, parents } = signedMessage.message;
        const id = signedMessage.id;
        this.log("inserting message %s at clock %d with parents %o", id, clock, parents);
        const result = await this.tree.write(async (txn) => {
            if (txn.has(signedMessage.key)) {
                return null;
            }
            return await this.apply(txn, signedMessage);
        });
        if (result !== null) {
            this.dispatchEvent(new CustomEvent("commit", { detail: result }));
        }
        return { id };
    }
    async apply(txn, signedMessage) {
        const { id, signature, message, key, value } = signedMessage;
        this.log.trace("applying %s %O", id, message);
        const messageId = new MessageId(id, key, message.clock);
        const parentIds = new MessageSet(message.parents.map(MessageId.encode));
        for (const parentId of parentIds) {
            if (!txn.has(parentId.key)) {
                throw new MissingParentError(parentId.id, id);
            }
        }
        const result = await this.#apply.apply(this, [signedMessage]);
        const hash = toString(hashEntry(key, value), "hex");
        const messageRecord = { id, data: signedMessage.value, hash, clock: message.clock };
        const effects = [
            { model: "$heads", operation: "set", value: { id } },
            { model: "$messages", operation: "set", value: messageRecord },
        ];
        const newHeads = [id];
        const oldHeads = await this.db.getAll("$heads");
        for (const head of oldHeads) {
            if (message.parents.includes(head.id)) {
                effects.push({ model: "$heads", operation: "delete", key: head.id });
            }
            else {
                newHeads.push(head.id);
            }
        }
        newHeads.sort();
        await this.ancestorIndex.indexAncestors(messageId, parentIds, effects);
        await this.db.apply(effects);
        txn.set(key, value);
        this.dispatchEvent(new CustomEvent("message", { detail: signedMessage }));
        const root = txn.getRoot();
        return { root, heads: newHeads, result };
    }
    async isAncestor(root, ancestor) {
        const ids = Array.isArray(root) ? root : [root];
        const visited = new MessageSet();
        for (const id of ids) {
            const isAncestor = await this.ancestorIndex.isAncestor(typeof id === "string" ? MessageId.encode(id) : id, typeof ancestor === "string" ? MessageId.encode(ancestor) : ancestor, visited);
            if (isAncestor) {
                return true;
            }
        }
        return false;
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(snapshot, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        let complete = true;
        let source = undefined;
        if (options.peer !== undefined) {
            source = { type: "sync", peer: options.peer };
        }
        let executionDuration = 0;
        await this.tree.read(async (txn) => {
            const driver = new sync.Driver(this.topic, snapshot, txn);
            try {
                for await (const keys of driver.sync()) {
                    const values = await snapshot.getValues(keys);
                    for (const [key, value] of zip(keys, values)) {
                        const signedMessage = this.decode(value, { source });
                        assert(equals(key, signedMessage.key), "invalid message key");
                        const executionStart = performance.now();
                        await this.insert(signedMessage);
                        executionDuration += performance.now() - executionStart;
                        messageCount++;
                    }
                }
            }
            catch (err) {
                if (err instanceof AbortError) {
                    complete = false;
                }
                else {
                    throw err;
                }
            }
        });
        const duration = Math.ceil(performance.now() - start);
        const peer = options.peer ?? "unknown";
        if (complete) {
            this.log("completed sync with %s (%dms total)", peer, duration);
        }
        else {
            this.log("aborted sync with %s (%dms total)", peer, duration);
        }
        this.log("applied %d messages in %dms", messageCount, Math.ceil(executionDuration));
        this.dispatchEvent(new CustomEvent("sync", { detail: { peer: options.peer, messageCount, duration } }));
        return { complete, messageCount };
    }
    serve(callback) {
        return this.tree.read((txn) => callback({
            getRoot: () => txn.getRoot(),
            getNode: (level, key) => txn.getNode(level, key),
            getChildren: (level, key) => txn.getChildren(level, key),
            getValues: async (keys) => {
                const values = [];
                const ids = keys.map(decodeId);
                const messageRecords = await this.db.getMany("$messages", ids);
                for (let i = 0; i < messageRecords.length; i++) {
                    const messageRecord = messageRecords[i];
                    if (messageRecord === null) {
                        throw new MessageNotFoundError(ids[i]);
                    }
                    values.push(messageRecord.data);
                }
                return values;
            },
        }));
    }
}

==> gossiplog-0.14.1-next.2.tgz: package/lib/AbstractGossipLog.js <==
import { TypedEventEmitter } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { equals, toString } from "uint8arrays";
import { hashEntry } from "@canvas-js/okra";
import { ed25519, prepareMessage } from "@canvas-js/signatures";
import { assert, zip } from "@canvas-js/utils";
import { AbortError, MessageNotFoundError, MissingParentError } from "@canvas-js/gossiplog/errors";
import * as sync from "@canvas-js/gossiplog/sync";
import target from "#target";
import { namespace, version } from "#migrations";
import { AncestorIndex } from "./AncestorIndex.js";
import { MessageSet } from "./MessageSet.js";
import { MessageId, decodeId, messageIdPattern } from "./MessageId.js";
import { SignedMessage, getNextClock } from "./SignedMessage.js";
import { gossiplogTopicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    db;
    tree;
    static namespace = namespace;
    static version = version;
    static schema = {
        $messages: {
            id: "primary",
            data: "bytes",
            hash: "string",
            clock: "integer",
            $indexes: ["clock"],
        },
        $heads: { id: "primary" },
        $replays: {
            timestamp: "primary",
            cursor: "string?",
            $indexes: ["cursor"],
        },
        ...AncestorIndex.schema,
    };
    topic;
    signer;
    controller = new AbortController();
    log;
    ancestorIndex;
    validatePayload;
    verifySignature;
    #apply;
    constructor(db, tree, init) {
        super();
        this.db = db;
        this.tree = tree;
        assert(gossiplogTopicPattern.test(init.topic), "invalid topic (must be of the form 'topic' or 'topic#hash', where topic matches [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.signer = init.signer ?? ed25519.create();
        this.#apply = init.apply;
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
        this.ancestorIndex = new AncestorIndex(db);
    }
    async initialize() {
        const replays = await this.db.query("$replays", {
            where: { cursor: { neq: null } },
            orderBy: { timestamp: "desc" },
        });
        if (replays.length > 0) {
            const [{ timestamp, cursor }, ...rest] = replays;
            assert(cursor !== null, "internal error - expected cursor !== null");
            this.log("found incomplete replay from %s at cursor %s", timestamp, cursor);
            // clear existing replay records
            await this.db.apply(rest.map(({ timestamp }) => ({
                model: "$replays",
                operation: "set",
                value: { timestamp, cursor: null },
            })));
            await this.#replay(timestamp, cursor);
        }
    }
    async close() {
        this.log("closing");
        this.controller.abort();
        await new Promise((resolve) => setTimeout(resolve, 0));
        await this.tree.close();
        await this.db.close();
    }
    async replay() {
        this.log("beginning replay");
        // clear existing replay records
        await this.db.query("$replays", { where: { cursor: { neq: null } } }).then(async (replays) => {
            const effects = replays.map((replay) => ({
                model: "$replays",
                operation: "set",
                value: { ...replay, cursor: null },
            }));
            await this.db.apply(effects);
        });
        const timestamp = new Date().toISOString();
        await this.db.set("$replays", { timestamp, cursor: null });
        for (const name of ["$heads", ...Object.keys(AncestorIndex.schema)]) {
            this.log("clearing model %s", name);
            await this.db.clear(name);
        }
        await this.tree.clear();
        return await this.#replay(timestamp);
    }
    async #replay(timestamp, cursor) {
        const pageSize = 128;
        const lowerBound = { gt: cursor };
        while (!this.controller.signal.aborted) {
            this.log("fetching new page");
            const results = await this.db.query("$messages", {
                orderBy: { id: "asc" },
                where: { id: lowerBound },
                limit: pageSize,
            });
            this.log("got new page of %d records", results.length);
            if (results.length === 0) {
                this.log("finished replaying");
                await this.db.set("$replays", { timestamp, cursor: null });
                return true;
            }
            let [{ id: cursor }] = results;
            await this.tree.write(async (txn) => {
                for (const { id, data } of results) {
                    this.log("replaying message %s", id);
                    const signedMessage = this.decode(data);
                    assert(signedMessage.id === id, "internal error - expected signedMessage.id === id");
                    await this.apply(txn, signedMessage);
                    cursor = id;
                    if (this.controller.signal.aborted) {
                        break;
                    }
                }
            });
            lowerBound.gt = cursor;
            await this.db.set("$replays", { timestamp, cursor });
        }
        this.log("replay incomplete");
        return false;
    }
    async connect(url, options = {}) {
        return await target.connect(this, url, options);
    }
    async listen(port, options = {}) {
        await target.listen(this, port, options);
    }
    async startLibp2p(config) {
        const libp2p = await target.startLibp2p(this, config);
        this.controller.signal.addEventListener("abort", () => libp2p.stop());
        return libp2p;
    }
    encode(signature, message, context = {}) {
        if (this.topic !== message.topic) {
            this.log.error("invalid message: %O", message);
            throw new Error(`cannot encode message: expected topic ${this.topic}, found topic ${message.topic}`);
        }
        const preparedMessage = prepareMessage(message);
        if (!this.validatePayload(preparedMessage.payload)) {
            this.log.error("invalid message: %O", preparedMessage);
            throw new Error(`error encoding message: invalid payload`);
        }
        return SignedMessage.encode(signature, preparedMessage, context);
    }
    decode(value, context = {}) {
        const signedMessage = SignedMessage.decode(value, context);
        const { topic, payload } = signedMessage.message;
        if (this.topic !== topic) {
            this.log.error("invalid message: %O", signedMessage.message);
            throw new Error(`cannot decode message: expected topic ${this.topic}, found topic ${topic}`);
        }
        if (!this.validatePayload(payload)) {
            this.log.error("invalid message: %O", signedMessage.message);
            throw new Error(`error decoding message: invalid payload`);
        }
        return signedMessage;
    }
    async getHeads() {
        const heads = await this.db.getAll("$heads");
        return new MessageSet(heads.map((head) => MessageId.encode(head.id)));
    }
    async getClock() {
        const heads = await this.db.getAll("$heads");
        const ids = heads.map(({ id }) => id);
        const clock = getNextClock(ids);
        return [clock, ids];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const records = await this.db.query("$messages", { select: { id: true }, where: { id } });
        return records.length > 0;
    }
    async get(id) {
        const record = await this.db.get("$messages", id);
        if (record === null) {
            return null;
        }
        else {
            return this.decode(record.data);
        }
    }
    async getMessages(range = {}) {
        const { reverse = false, limit, ...where } = range;
        const records = await this.db.query("$messages", {
            where: { id: where },
            select: { id: true, data: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
        return records.map(({ data }) => this.decode(data));
    }
    async *iterate(range = {}) {
        const { reverse = false, limit, ...where } = range;
        for await (const row of this.db.iterate("$messages", {
            where: { id: where },
            select: { id: true, data: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        })) {
            yield this.decode(row.data);
        }
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The current concurrent heads of the local log are used as parents.
     */
    async append(payload, { signer = this.signer } = {}) {
        let root = null;
        let heads = null;
        let result = undefined;
        const signedMessage = await this.tree.write(async (txn) => {
            const [clock, parents] = await this.getClock();
            const message = prepareMessage({ topic: this.topic, clock, parents, payload });
            const signature = await signer.sign(message);
            const signedMessage = this.encode(signature, message);
            this.log("appending message %s at clock %d with parents %o", signedMessage.id, clock, parents);
            const applyResult = await this.apply(txn, signedMessage);
            root = applyResult.root;
            heads = applyResult.heads;
            result = applyResult.result;
            return signedMessage;
        });
        assert(root !== null && heads !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root, heads } }));
        signedMessage.result = result;
        return signedMessage;
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(signedMessage) {
        const { message, signature } = signedMessage;
        assert(message.topic === this.topic, `expected message.topic === this.topic`);
        await this.verifySignature(signature, message);
        const { clock, parents } = signedMessage.message;
        const id = signedMessage.id;
        this.log("inserting message %s at clock %d with parents %o", id, clock, parents);
        const result = await this.tree.write(async (txn) => {
            if (txn.has(signedMessage.key)) {
                return null;
            }
            return await this.apply(txn, signedMessage);
        });
        if (result !== null) {
            this.dispatchEvent(new CustomEvent("commit", { detail: result }));
        }
        return { id };
    }
    async apply(txn, signedMessage) {
        const { id, signature, message, key, value } = signedMessage;
        this.log.trace("applying %s %O", id, message);
        const messageId = new MessageId(id, key, message.clock);
        const parentIds = new MessageSet(message.parents.map(MessageId.encode));
        for (const parentId of parentIds) {
            if (!txn.has(parentId.key)) {
                throw new MissingParentError(parentId.id, id);
            }
        }
        const result = await this.#apply.apply(this, [signedMessage]);
        const hash = toString(hashEntry(key, value), "hex");
        const messageRecord = { id, data: signedMessage.value, hash, clock: message.clock };
        const effects = [
            { model: "$heads", operation: "set", value: { id } },
            { model: "$messages", operation: "set", value: messageRecord },
        ];
        const newHeads = [id];
        const oldHeads = await this.db.getAll("$heads");
        for (const head of oldHeads) {
            if (message.parents.includes(head.id)) {
                effects.push({ model: "$heads", operation: "delete", key: head.id });
            }
            else {
                newHeads.push(head.id);
            }
        }
        newHeads.sort();
        await this.ancestorIndex.indexAncestors(messageId, parentIds, effects);
        await this.db.apply(effects);
        txn.set(key, value);
        this.dispatchEvent(new CustomEvent("message", { detail: signedMessage }));
        const root = txn.getRoot();
        return { root, heads: newHeads, result };
    }
    async isAncestor(root, ancestor) {
        const visited = new MessageSet();
        const ids = root instanceof MessageSet ? root : Array.isArray(root) ? root : [root];
        const ancestorId = typeof ancestor === "string" ? MessageId.encode(ancestor) : ancestor;
        for (const id of ids) {
            const rootId = typeof id === "string" ? MessageId.encode(id) : id;
            this.log.trace("isAncestor %s ?-> %s", rootId.clock, rootId.id, ancestorId.clock, ancestorId.id);
            const isAncestor = await this.ancestorIndex.isAncestor(rootId, ancestorId, visited);
            if (isAncestor) {
                return true;
            }
        }
        return false;
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(snapshot, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        let complete = true;
        let source = undefined;
        if (options.peer !== undefined) {
            source = { type: "sync", peer: options.peer };
        }
        let executionDuration = 0;
        await this.tree.read(async (txn) => {
            const driver = new sync.Driver(this.topic, snapshot, txn);
            try {
                for await (const keys of driver.sync()) {
                    const values = await snapshot.getValues(keys);
                    for (const [key, value] of zip(keys, values)) {
                        const signedMessage = this.decode(value, { source });
                        assert(equals(key, signedMessage.key), "invalid message key");
                        const executionStart = performance.now();
                        await this.insert(signedMessage);
                        executionDuration += performance.now() - executionStart;
                        messageCount++;
                    }
                }
            }
            catch (err) {
                if (err instanceof AbortError) {
                    complete = false;
                }
                else {
                    throw err;
                }
            }
        });
        const duration = Math.ceil(performance.now() - start);
        const peer = options.peer ?? "unknown";
        if (complete) {
            this.log("completed sync with %s (%dms total)", peer, duration);
        }
        else {
            this.log("aborted sync with %s (%dms total)", peer, duration);
        }
        this.log("applied %d messages in %dms", messageCount, Math.ceil(executionDuration));
        this.dispatchEvent(new CustomEvent("sync", { detail: { peer: options.peer, messageCount, duration } }));
        return { complete, messageCount };
    }
    serve(callback) {
        return this.tree.read((txn) => callback({
            getRoot: () => txn.getRoot(),
            getNode: (level, key) => txn.getNode(level, key),
            getChildren: (level, key) => txn.getChildren(level, key),
            getValues: async (keys) => {
                const values = [];
                const ids = keys.map(decodeId);
                const messageRecords = await this.db.getMany("$messages", ids);
                for (let i = 0; i < messageRecords.length; i++) {
                    const messageRecord = messageRecords[i];
                    if (messageRecord === null) {
                        throw new MessageNotFoundError(ids[i]);
                    }
                    values.push(messageRecord.data);
                }
                return values;
            },
        }));
    }
}
